import sys
import catalogue
import confection
from confection import Config, ConfigValidationError
from typing import TypeVar, Callable

_DIn = TypeVar("_DIn")

# Use typing_extensions for Python versions < 3.8
if sys.version_info < (3, 8):
    from typing_extensions import Protocol, Literal
else:
    from typing import Protocol, Literal  # noqa: F401


class Decorator(Protocol):
    """Protocol to mark a function as returning its child with identical signature."""

    def __call__(self, name: str) -> Callable[[_DIn], _DIn]:
        ...


class registry(confection.registry):
    @classmethod
    def create(cls, registry_name: str, entry_points: bool = False) -> None:
        """Create a new custom registry."""
        if hasattr(cls, registry_name):
            # raise ValueError(f"Registry '{registry_name}' already exists")
            return
        reg: Decorator = catalogue.create(
            "torch_snippets", registry_name, entry_points=entry_points
        )
        setattr(cls, registry_name, reg)


__all__ = ["Config", "registry", "ConfigValidationError"]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/charts.ipynb.

# %% auto 0
__all__ = ['alt', 'Chart', 'CM', 'confusion_matrix']

# %% ../nbs/charts.ipynb 2
import altair as alt
from altair import Chart

alt = alt
Chart = Chart
from .loader import *

# %% ../nbs/charts.ipynb 5
def confusion_matrix(df=None, truth=None, pred=None, mapping=None, save_to=None):
    if df is None:
        df = pd.DataFrame({"truth": truth, "pred": pred})
        truth = "truth"
        pred = "pred"
    DF = df[[truth, pred]]

    try:
        from sklearn.metrics import classification_report

        print(classification_report(DF[truth], DF[pred]))
    except:
        logger.info("Skipping Report")
    df = DF.pivot_table(index=truth, columns=pred, aggfunc=len).reset_index()
    df = df.melt(id_vars=truth, var_name=pred, value_name="value")
    sz = 450 if len(DF[truth].unique()) > 4 else 250

    base = alt.Chart(df, height=sz, width=sz)
    hm = base.mark_rect().encode(
        x=f"{pred}:O",
        y=f"{truth}:O",
        color=alt.Color("value:Q", scale=alt.Scale(scheme="lightorange"), legend=None),
    )

    tx = base.mark_text(baseline="middle").encode(
        x=f"{pred}:O",
        y=f"{truth}:O",
        text="value:Q",
        color=alt.condition(
            alt.datum.value > 0,
            alt.value("black"),
            alt.value("rgba(0, 0, 0, 0)"),  # Transparent color for value 0
        ),
    )

    if save_to is not None:
        assert str(save_to).endswith("html"), "Can only save in html format"
        (hm + tx).save(save_to, format="html")
        Info(f"Saved chart at {save_to}")
    else:
        return hm + tx


CM = confusion_matrix



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/capsule.ipynb.

# %% auto 0
__all__ = ['to', 'train', 'validate', 'predict', 'Capsule']

# %% ../../nbs/capsule.ipynb 2
from functools import wraps
from ..loader import *
from ..torch_loader import *
from ..paths import loaddill, dumpdill, makedir, parent
from ..load_defaults import exists
from ..markup import AttrDict

try:
    import mmcv
    from mmcv.parallel.data_container import DataContainer
except ImportError:
    DataContainer = None

# %% ../../nbs/capsule.ipynb 3
def to(item, device):
    if item is None:
        return None
    elif isinstance(item, (torch.Tensor, nn.Module)):
        return item.to(device)
    elif isinstance(item, (AttrDict, dict)):
        return type(item)({k: to(v, device) for k, v in item.items()})
    elif isinstance(item, (list, tuple)):
        return type(item([to(_item, device) for _item in item]))
    elif DataContainer is not None and isinstance(item, DataContainer):
        return DataContainer([to(_item, device) for _item in item.data])
    else:
        # logger.warning(f"function is not implemented for {type(item)}")
        return item
        raise NotImplementedError(f"function is not implemented for {type(item)}")


def train(train_function):
    @wraps(train_function)
    def _train_batch(self, *args, **kwargs):
        args = self.before_train_batch(args)
        kwargs = self.before_train_batch(kwargs)
        outputs = train_function(self, *args, **kwargs)
        outputs = self.after_train_batch(outputs)
        assert isinstance(outputs, dict)
        return outputs

    return _train_batch


@torch.no_grad()
def validate(validation_function):
    @wraps(validation_function)
    def _validate_batch(self, *args, **kwargs):
        args = self.before_validate_batch(args)
        kwargs = self.before_validate_batch(kwargs)
        outputs = validation_function(self, *args, **kwargs)
        outputs = self.after_validate_batch(outputs)
        assert isinstance(outputs, dict)
        return outputs

    return _validate_batch


@torch.no_grad()
def predict(predict_function):
    @wraps(predict_function)
    def _predict(self, *args, **kwargs):
        args = self.before_predict(args)
        kwargs = self.before_predict(kwargs)
        outputs = predict_function(self, *args, **kwargs)
        return outputs

    return _predict

# %% ../../nbs/capsule.ipynb 4
class Capsule(nn.Module):
    """
    Inherit from Capsule and train with fewer lines of code
    """

    def __init__(self, report=None):
        super().__init__()
        if report is not None:
            self.report = loaddill(report)

    # Train Utils
    def before_train_batch(self, data):
        self.train()
        self.optimizer.zero_grad()
        data = to(data, getattr(self, "device", "cuda"))
        return data

    def after_train_batch(self, outputs):
        outputs["loss"].backward()
        self.optimizer.step()
        return outputs

    # Validation Utils
    def before_validate_batch(self, data):
        self.eval()
        data = to(data, getattr(self, "device", "cuda"))
        return data

    def after_validate_batch(self, outputs):
        return outputs

    def before_predict(self, data):
        self.eval()
        data = to(data, getattr(self, "device", "cuda"))
        return data

    def after_predict(self, outputs):
        return outputs

    def load(self, weights_path=None, device="cpu"):
        if weights_path:
            load_torch_model_weights_to(self, weights_path, device=device)
        try:
            weights_path = weights_path + ".report"
            self.report = loaddill(weights_path)
        except:
            pass

    def save(self, save_to):
        if not exists(parent(save_to)):
            makedir(parent(save_to), prompt=False)
        save_torch_model_weights_from(self, save_to)
        if hasattr(self, "report"):
            save_to = save_to + ".report"
            dumpdill(self.report, save_to)

    # Fit function
    def fit(
        self,
        trn_dl=None,
        val_dl=None,
        num_epochs=1,
        lr=None,
        device="cuda",
        save_to=None,
        print_every=None,
        print_total=None,
        show_final_plot=True,
        **kwargs,
    ):
        if print_total:
            print_every = num_epochs // print_total

        if lr:
            for group in self.optimizer.param_groups:
                group["lr"] = lr
            Info(f"Learning Rate: {lr}")

        if not hasattr(self, "report"):
            self.report = Report(num_epochs, **kwargs)
        else:
            self.report = Report(num_epochs, old_report=self.report, **kwargs)

        self.device = device
        to(self, self.device)

        try:
            for epoch in range(num_epochs):
                self.report.n_epochs = num_epochs
                if trn_dl is not None:
                    N = len(trn_dl)
                    for ix, data in enumerate(trn_dl):
                        loss = self.train_batch(data)
                        self.report.record(pos=(epoch + (ix + 1) / N), **loss, end="\r")
                if val_dl is not None:
                    self.evaluate(
                        val_dl, report=self.report, device=device, epoch=epoch
                    )
                if (print_every and ((epoch + 1) % print_every == 0)) or epoch == 0:
                    self.report.report_avgs(epoch + 1)
                else:
                    self.report.report_avgs(epoch + 1, end="\r")
        except KeyboardInterrupt:
            pass

        if show_final_plot:
            self.report.plot(log=True, smooth=0)
        if save_to:
            self.save(save_to)
        self.report.finish_run(**kwargs)

    def evaluate(
        self, val_dl, report=None, device="cuda", epoch=None, show_report=False
    ):
        if report is None:
            report = Report(1)
            epoch = 0
            report_averages = True
        else:
            epoch = epoch
            report_averages = False
        self.device = device
        to(self, self.device)

        N = len(val_dl)
        for ix, data in enumerate(val_dl):
            loss = self.validate_batch(data)
            report.record(pos=(epoch + (ix + 1) / N), **loss, end="\r")

        avgs = report.report_avgs(1) if report_averages else None
        if show_report:
            report.plot(log=True, smooth=3)
        return avgs

    @train
    def train_batch(self, data):
        x, y = data
        _y = self(x)
        loss = self.criterion(_y, y)
        return {"loss": loss}

    @validate
    def validate_batch(self, data):
        x, y = data
        output = self(x)
        loss = self.criterion(output, y)
        return {"loss": loss}

    @predict
    def predict_batch(self, data):
        x, _ = data
        outputs = self(x)
        return outputs



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/misc.ipynb.

# %% auto 0
__all__ = ['Timer', 'track2', 'timeit', 'io']

# %% ../nbs/misc.ipynb 2
import time
from .logger import Debug
from .inspector import inspect
from fastcore.basics import ifnone

# %% ../nbs/misc.ipynb 3
class Timer:
    def __init__(self, N, smooth=True):
        "print elapsed time every iteration and print out remaining time"
        "assumes this timer is called exactly N times or less"
        self.tok = self.start = time.time()
        self.N = N
        self.ix = 0
        self.smooth = smooth

    def __call__(self, ix=None, info=None):
        ix = self.ix if ix is None else ix
        info = "" if info is None else f"{info}\t"
        tik = time.time()
        elapsed = tik - self.start
        ielapsed = tik - self.tok
        ispeed = ielapsed

        iunit = "s/iter"
        if ispeed < 1:
            ispeed = 1 / ispeed
            iunit = "iters/s"

        iremaining = (self.N - (ix + 1)) * ielapsed
        iestimate = iremaining + elapsed

        print(
            f"{info}{ix+1}/{self.N} ({elapsed:.2f}s - {iremaining:.2f}s remaining - {ispeed:.2f} {iunit})"
            + " " * 10,
            end="\r",
        )
        self.ix += 1
        self.tok = tik


def track2(iterable, *, total=None):
    try:
        total = ifnone(total, len(iterable))
    except:
        ...
    timer = Timer(total)
    for item in iterable:
        info = yield item
        timer(info=info)
        if info is not None:
            yield  # Just to ensure the send operation stops

# %% ../nbs/misc.ipynb 10
def timeit(func):
    def inner(*args, **kwargs):
        s = time.time()
        o = func(*args, **kwargs)
        Debug(f"{time.time() - s:.2f} seconds to execute `{func.__name__}`")
        return o

    return inner


def io(func):
    def inner(*args, **kwargs):
        s = time.time()
        o = func(*args, **kwargs)
        Debug(f"Args: {inspect(args)}\nKWargs: {inspect(kwargs)}\nOutput: {inspect(o)}")
        return o

    return inner



from IPython.display import clear_output, display
from matplotlib import rc
from matplotlib.animation import FuncAnimation

from torch_snippets import *

rc("animation", html="jshtml")


def get_sz(w, sz=None):
    if sz is None:
        if w < 129:
            sz = 2
        elif w < 1025:
            sz = 5
        else:
            sz = 10
    if isinstance(sz, int):
        sz = (sz, sz)
    return sz


def denormalize_image(img):
    Debug(f"DeNormalizing...")
    if img.min() < 0:
        img = (img - img.min()) / (img.max() - img.min())
    img = (img * 255).astype(np.uint8)
    return img


def animate(frames, sz: int = None):
    if isinstance(frames, torch.Tensor):
        frames = frames.cpu().numpy()
    Info(f"Animating {torch.Tensor(frames)}")

    # if isinstance(frames, list):
    #     frames = np.array(frames)
    if isinstance(frames, np.ndarray):
        if len(frames.shape) == 5:
            # assuming BXXXX
            Warn(f"Received a 5D tensor. Animating only the first one in the batch")
            frames = frames[0]

        if len(frames.shape) == 4:
            if frames.shape[1] == 3:
                Debug("assuming NTHW")
                frames = frames.transpose(0, 2, 3, 1)  # Convert to THWC format
            if frames.shape[0] == 3:
                Debug("assuming TNHW")
                frames = frames.transpose(1, 2, 3, 0)  # Convert to THWC format

        # if frames.min() < 0 or frames.max() > 1:
        #     frames = denormalize_image(frames)

    w = frames[0].shape[1]
    h, w = frames.shape[1:3]
    sz = sz * w // h, sz

    def init():
        ax.set_xlim(0, 10)
        ax.set_ylim(-1, 1)
        ax.margins(x=0, y=0)
        return []

    def update(frame):
        ax.clear()
        img = frames[frame]
        ax.imshow(img)
        ax.axis("off")
        return []

    fig, ax = plt.subplots(figsize=sz)
    ax.set_adjustable("box")
    fig.subplots_adjust(
        left=0.01, bottom=0.01, right=0.99, top=0.99, wspace=None, hspace=None
    )
    ani = FuncAnimation(fig, update, frames=len(frames), init_func=init, blit=True)
    plt.close()
    return ani



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/decorators.ipynb.

# %% auto 0
__all__ = ['timeit', 'io', 'check_kwargs_not_none']

# %% ../nbs/decorators.ipynb 2
from functools import wraps
from .inspector import inspect
from .logger import Info
import time

# %% ../nbs/decorators.ipynb 3
def timeit(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        output = func(*args, **kwargs)
        end = time.time()
        Info(f"{func.__name__} took {end-start:.2f} seconds to execute")
        return output

    return wrapper


def io(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if len(args) != 0:
            inspect(args, names=["inputs:args"])
        if kwargs != {}:
            inspect(kwargs, names=["inputs:kwargs"])
        output = func(*args, **kwargs)
        inspect(output, names=["outputs"])
        return output

    return wrapper


def check_kwargs_not_none(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        for key, value in kwargs.items():
            if value is None:
                raise ValueError(f"Input argument '{key}' cannot be None")
        return func(*args, **kwargs)

    return wrapper



__version__ = "0.526"
from .loader import *
from .paths import *
from .markup import *
from .inspector import *
from .load_defaults import *
from .pdf_loader import PDF
from .logger import *
from .markup2 import AD

try:
    from .ipython import *
except:
    ...

from .decorators import *
from .misc import *
from .dates import *

try:
    from .torch_loader import *
except Exception as e:
    ...



"v0.15.41"
"""Todo - v0.15.5
- [ ] accumulate lists of tensors
- [ ] make plot_epochs even faster
"""

__all__ = [
    "torch",
    "th",
    "torchvision",
    "T",
    "transforms",
    "nn",
    "np",
    "F",
    "Dataset",
    "DataLoader",
    "optim",
    "Report",
    "Reshape",
    "Permute",
    "device",
    "save_torch_model_weights_from",
    "load_torch_model_weights_to",
    "detach",
    "cat_with_padding",
]

import re
import time
from collections import defaultdict, namedtuple
from itertools import dropwhile, takewhile
from typing import List, Union

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch as th
import torch.nn as nn
import torchvision
import torchvision.transforms as T
from loguru import logger
from torch import optim
from torch.nn import functional as F
from torch.utils.data import DataLoader, Dataset
from torchvision import transforms

from .paths import makedir, os

device = "cuda" if torch.cuda.is_available() else "cpu"

try:
    import wandb
except ImportError:
    pass

try:
    from mlflow_extend import mlflow
except ImportError:
    mlflow = None


class Reshape(nn.Module):
    def __init__(self, *shape):
        super().__init__()
        self.shape = shape

    def forward(self, x):
        return x.view(*self.shape)


class Permute(nn.Module):
    def __init__(self, *order):
        super().__init__()
        self.order = order

    def forward(self, x):
        return x.permute(*self.order)


metric = namedtuple("metric", "pos,val".split(","))


def process(v):
    if isinstance(v, torch.Tensor):
        if v.numel() == 1:
            return v.item()
    return v


def info(report, precision):
    report = {k: process(v) for k, v in report.items()}
    return "  ".join([f"{k}: {v:.{precision}f}" for k, v in report.items()])


def to_np(x):
    if isinstance(x, torch.Tensor):
        return float(x.detach().cpu().numpy())
    else:
        return x


class Report:
    def __init__(
        self,
        n_epochs=None,
        precision=3,
        old_report=None,
        wandb_project=None,
        hyper_parameters=None,
        **kwargs,
    ):
        self.start = time.time()
        self.n_epochs = n_epochs
        self.precision = precision
        self.completed_epochs = -1
        self.logged = set()
        self.set_external_logging(wandb_project, hyper_parameters)
        if old_report:
            self.prepend(old_report)

    def reset_time(self):
        self.start = time.time()

    def prepend(self, old_report):
        for k in old_report.logged:
            self.logged.append(k)
            last = getattr(old_report, k)[-1].pos
            setattr(self, k, [])
            for m in getattr(old_report, k):
                getattr(self, k).append(metric(m.pos - last, to_np(m.val)))

    def record(self, pos, **metrics):
        metrics = {k: to_np(v) for k, v in metrics.items()}
        for k, v in metrics.items():
            if k in ["end", "pos"]:
                continue
            if hasattr(self, k):
                getattr(self, k).append(metric(pos, to_np(v)))
            else:
                setattr(self, k, [])
                getattr(self, k).append(metric(pos, to_np(v)))
                self.logged.add(k)

        if not any(["val" in key for key in metrics.keys()]):
            key = "train_step"
            step = self.train_step
        else:
            key = "validation_step"
            step = self.validation_step

        self.report_metrics(pos, **metrics)

        metrics = {k: v for k, v in metrics.items() if not isinstance(v, str)}
        if self.wandb_logging:
            wandb.log({**metrics, key: step})
        if self.mlflow_logging:
            mlflow.log_metrics(metrics, step=step)

        if key == "train_step":
            self.train_step += 1
        else:
            self.validation_step += 1

    def plot(self, keys: Union[List, str] = None, smooth=0, ax=None, **kwargs):
        _show = True if ax is None else False
        if ax is None:
            sz = 8, 6
            fig, ax = plt.subplots(figsize=kwargs.get("figsize", sz))

        keys = self.logged if keys is None else keys
        if isinstance(keys, str):
            key_pattern = keys
            keys = [key for key in self.logged if re.search(key_pattern, key)]

        for k in keys:
            xs, ys = list(zip(*getattr(self, k)))
            if smooth:
                ys = moving_average(np.array(ys), smooth)

            if "val" in k:
                _type = "--"
            elif "test" in k:
                _type = ":"
            else:
                _type = "-"

            ax.plot(xs, ys, _type, label=k)
        ax.grid(True)
        ax.set_xlabel("Epochs")
        ax.set_ylabel("Metrics")
        ax.set_title(
            kwargs.get("title", None), fontdict=kwargs.get("fontdict", {"size": 20})
        )
        if kwargs.get("log", False):
            ax.semilogy()
        ax.legend()
        if _show:
            plt.show()

    def history(self, k):
        return [v for _, v in getattr(self, k)]

    def plot_epochs(self, keys: list = None, ax=None, **kwargs):
        _show = True if ax is None else False
        if ax is None:
            sz = 8, 6
            fig, ax = plt.subplots(figsize=kwargs.get("figsize", sz))
        avgs = defaultdict(list)
        keys = self.logged if keys is None else keys

        from tqdm import trange

        if isinstance(keys, str):
            key_pattern = keys
            keys = [key for key in self.logged if re.search(key_pattern, key)]
        xs = []
        for epoch in trange(-100, self.n_epochs + 1):
            for k in keys:
                items = takewhile(
                    lambda x: epoch - 1 <= x.pos < epoch,
                    dropwhile(
                        lambda x: (epoch - 1 > x.pos or x.pos > epoch), getattr(self, k)
                    ),
                )
                items = list(items)
                if items == []:
                    continue
                xs.append(epoch)
                avgs[k].append(np.mean([v for pos, v in items]))
            xs = sorted(set(xs))

        for k in avgs:
            if "val" in k:
                _type = "--"
            elif "test" in k:
                _type = ":"
            else:
                _type = "-"
            if len(avgs[k]) != len(xs):
                logger.info(
                    f"metric {k} was not fully recorded. Plotting final epochs using last recorded value"
                )
                avgs[k].extend([avgs[k][-1]] * (len(xs) - len(avgs[k])))
            ax.plot(
                xs,
                avgs[k],
                _type,
                label=k,
            )
        ax.grid(True)
        ax.set_xlabel("Epochs")
        ax.set_ylabel("Metrics")
        ax.set_title(
            kwargs.get("title", None), fontdict=kwargs.get("fontdict", {"size": 20})
        )
        if kwargs.get("log", False):
            ax.semilogy()
        plt.legend()
        if _show:
            plt.show()

    def report_avgs(self, epoch, return_avgs=True, end="\n"):
        avgs = {}
        for k in self.logged:
            avgs[k] = np.mean(
                [v for pos, v in getattr(self, k) if epoch - 1 <= pos < epoch]
            )
        self.report_metrics(epoch, end=end, **avgs)
        avgs = {f"epoch_{k}": v for k, v in avgs.items()}
        if self.wandb_logging:
            wandb.log({**avgs, "epoch": epoch})
        if self.mlflow_logging:
            mlflow.log_metrics(avgs, step=epoch)
        if return_avgs:
            return avgs

    def report_metrics(self, pos, **report):
        """Report training and validation metrics
        Required variables to be initialized before calling this function:
        1. start (time.time())
        2. n_epochs (int)

        Special kwargs:
        1. end - line ending after print (default newline)
        2. log - prefix info before printing summary

        Special argument:
        1. pos - position in training/testing process - float between 0 - n_epochs

        Usage:
        report_metrics(pos=1.42, train_loss=train_loss, validation_loss=validation_loss, ... )
        where each kwarg is a float
        """
        elapsed = time.time() - self.start
        end = report.pop("end", "\n")
        log = report.pop("log", "")
        log = log + ": " if log != "" else log
        elapsed = "  ({:.2f}s - {:.2f}s remaining)".format(
            time.time() - self.start, ((self.n_epochs - pos) / pos) * elapsed
        )
        current_iteration = f"EPOCH: {pos:.3f}  "
        if end == "\r":
            print(
                f"\r{log}{current_iteration}{info(report, self.precision)}{elapsed}",
                end="",
            )
        else:
            print(
                f"\r{log}{current_iteration}{info(report, self.precision)}{elapsed}",
                end=end,
            )

    def set_external_logging(self, project=None, hyper_parameters=None):
        if project is not None:
            if "/" in project:
                project, name = project.split("/")
            self.wandb_logging = True
            wandb.init(project=project, name=name, config=hyper_parameters)
        else:
            self.wandb_logging = False

        if mlflow and mlflow.active_run():
            self.mlflow_logging = True
        else:
            self.mlflow_logging = False

        self.train_step = 0
        self.validation_step = 0

    def finish_run(self, **kwargs):
        if not kwargs.get("do_not_finish_wandb", False):
            if self.wandb_logging:
                wandb.finish()


try:
    import pytorch_lightning as pl
    from pytorch_lightning.callbacks.progress import ProgressBarBase

    class LightningReport(ProgressBarBase):
        def __init__(
            self,
            epochs,
            print_every=None,
            print_total=None,
            precision=4,
            old_report=None,
        ):
            super().__init__()
            self.enable = True
            self.epoch_ix = 0
            _report = old_report.report if old_report is not None else None
            self.report = Report(epochs, precision, _report)
            if print_every is not None:
                self.print_every = print_every
            else:
                if print_total is None:
                    if epochs < 11:
                        self.print_every = 1
                else:
                    self.print_every = epochs // print_total

        def disable(self):
            self.enable = False

        def on_epoch_end(self, trainer, pl_module):
            self.epoch_ix += 1
            if self.epoch_ix % self.print_every == 0:
                self.report.report_avgs(self.epoch_ix)

        def prefix_dict_keys(self, prefix, _dict):
            _d = {}
            for k in _dict:
                _d[f"{prefix}_{k}"] = _dict[k]
            return _d

        def on_train_batch_end(
            self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
        ):
            super().on_train_batch_end(
                trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
            )
            if isinstance(outputs, list):
                loss = float(trainer.progress_bar_dict["loss"])
                outputs = outputs[0][0]["extra"]
                outputs["loss"] = loss
            outputs = self.prefix_dict_keys("trn", outputs)
            self.px = self.epoch_ix + ((1 + batch_idx) / self.total_train_batches)
            self.report.record(self.px, end="\r", **outputs)

        def on_validation_batch_end(
            self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
        ):
            super().on_train_batch_end(
                trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
            )
            loss = float(trainer.progress_bar_dict["loss"])
            self.px = self.epoch_ix + ((1 + batch_idx) / self.total_val_batches)
            outputs = self.prefix_dict_keys("val", outputs)
            self.report.record(self.px, end="\r", **outputs)

        def __getattr__(self, attr, **kwargs):
            return getattr(self.report, attr, **kwargs)

    __all__ += ["LightningReport", "pl"]
except Exception as e:
    ...


def moving_average(a, n=3):
    b = np.zeros_like(a)
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    _n = len(b) - n
    b[-_n - 1 :] = ret[(n - 1) :] / n
    return b


def save_torch_model_weights_from(model, fpath):
    "from model to fpath"
    torch.save(model.state_dict(), fpath)
    fsize = os.path.getsize(fpath) >> 20
    logger.opt(depth=1).log("INFO", f"Saved weights of size ~{fsize} MB to {fpath}")


def load_torch_model_weights_to(model, fpath, device=None):
    "to model from fpath"
    if not device:
        model.load_state_dict(torch.load(fpath))
    else:
        model.load_state_dict(torch.load(fpath, map_location=device))
    logger.opt(depth=1).log("INFO", f"Loaded weights from {fpath} to given model")


def detach(i):
    if isinstance(i, dict):
        for k, v in i.items():
            i[k] = detach(v)
        return i
    elif isinstance(i, (list, tuple)):
        return [detach(j) for j in i]
    else:
        return i.cpu().detach()


def cat_with_padding(tensors, mode="constant", value=-100):
    """
    Concatenates a list of tensors with padding to make them compatible along 0th dimension.

    Args:
        tensors (list of torch.Tensor): List of tensors to be concatenated.
        mode (str, optional): The padding mode. Default is "constant".
        value (int, optional): The padding value for "constant" mode. Default is -100.

    Returns:
        torch.Tensor: Concatenated tensor with padding to match dimensions.

    Note:
        - All tensors should have the same dimension except for the stacking dimension.
        - Padding is added to match the dimensions of the largest tensor in the specified stacking dimension.
    """
    assert all(
        [tensors[0].ndim == t.ndim for t in tensors[1:]]
    ), "All tensors should have the same number of dimensions"
    sizes = [(t.size()[1:]) for t in tensors]
    full_size = [max(s) for s in list(zip(*sizes))]
    if full_size == []:
        # Given tensors are just scalars
        return torch.stack(tensors)

    def make_padding(current_size, target_size):
        o = ()
        for curr_d, targ_d in zip(current_size[::-1], target_size[::-1]):
            diff_d = targ_d - curr_d
            o = o + (0, diff_d)
        return o

    out = torch.cat(
        [
            F.pad(x, make_padding(x.size()[1:], full_size), mode=mode, value=value)
            for x in tensors
        ],
        dim=0,
    )
    return out


"v0.15.41"
"""Todo - v0.15.5
- [ ] accumulate lists of tensors
- [ ] make plot_epochs even faster
"""

__all__ = [
    "torch",
    "th",
    "torchvision",
    "T",
    "transforms",
    "nn",
    "np",
    "F",
    "Dataset",
    "DataLoader",
    "optim",
    "Report",
    "Reshape",
    "Permute",
    "device",
    "save_torch_model_weights_from",
    "load_torch_model_weights_to",
    "detach",
    "cat_with_padding",
    "clean_gpu_mem",
    "get_latest_checkpoint",
]

import gc
import re
import time
import traceback
from collections import defaultdict, namedtuple
from itertools import dropwhile, takewhile
from typing import List, Union

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch as th
import torch.nn as nn
import torchvision
import torchvision.transforms as T
from loguru import logger
from torch import optim
from torch.nn import functional as F
from torch.utils.data import DataLoader, Dataset
from torchvision import transforms

from .paths import makedir, os

device = "cuda" if torch.cuda.is_available() else "cpu"

try:
    import wandb
except ImportError:
    pass

try:
    from mlflow_extend import mlflow
except ImportError:
    mlflow = None


class Reshape(nn.Module):
    def __init__(self, *shape):
        super().__init__()
        self.shape = shape

    def forward(self, x):
        return x.view(*self.shape)


class Permute(nn.Module):
    def __init__(self, *order):
        super().__init__()
        self.order = order

    def forward(self, x):
        return x.permute(*self.order)


metric = namedtuple("metric", "pos,val".split(","))


def process(v):
    if isinstance(v, torch.Tensor):
        if v.numel() == 1:
            return v.item()
    return v


def info(report, precision):
    report = {k: process(v) for k, v in report.items()}
    return "  ".join([f"{k}: {v:.{precision}f}" for k, v in report.items()])


def to_np(x):
    if isinstance(x, torch.Tensor):
        return float(x.detach().cpu().numpy())
    else:
        return x


class Report:
    def __init__(
        self,
        n_epochs=None,
        precision=3,
        old_report=None,
        wandb_project=None,
        hyper_parameters=None,
        **kwargs,
    ):
        self.start = time.time()
        self.n_epochs = n_epochs
        self.precision = precision
        self.completed_epochs = -1
        self.logged = set()
        self.set_external_logging(wandb_project, hyper_parameters)
        if old_report:
            self.prepend(old_report)

    def reset_time(self):
        self.start = time.time()

    def prepend(self, old_report):
        for k in old_report.logged:
            self.logged.add(k)
            last = getattr(old_report, k)[-1].pos
            setattr(self, k, [])
            for m in getattr(old_report, k):
                getattr(self, k).append(metric(m.pos - last, to_np(m.val)))

    def record(self, pos, **metrics):
        metrics = {k: to_np(v) for k, v in metrics.items()}
        for k, v in metrics.items():
            if k in ["end", "pos"]:
                continue
            if hasattr(self, k):
                getattr(self, k).append(metric(pos, to_np(v)))
            else:
                setattr(self, k, [])
                getattr(self, k).append(metric(pos, to_np(v)))
                self.logged.add(k)

        if not any(["val" in key for key in metrics.keys()]):
            key = "train_step"
            step = self.train_step
        else:
            key = "validation_step"
            step = self.validation_step

        self.report_metrics(pos, **metrics)

        metrics = {k: v for k, v in metrics.items() if not isinstance(v, str)}
        if self.wandb_logging:
            wandb.log({**metrics, key: step})
        if self.mlflow_logging:
            mlflow.log_metrics(metrics, step=step)

        if key == "train_step":
            self.train_step += 1
        else:
            self.validation_step += 1

    def plot(self, keys: Union[List, str] = None, smooth=0, ax=None, **kwargs):
        _show = True if ax is None else False
        if ax is None:
            sz = 8, 6
            fig, ax = plt.subplots(figsize=kwargs.get("figsize", sz))

        keys = self.logged if keys is None else keys
        if isinstance(keys, str):
            key_pattern = keys
            keys = [key for key in self.logged if re.search(key_pattern, key)]

        for k in keys:
            xs, ys = list(zip(*getattr(self, k)))
            if smooth:
                ys = moving_average(np.array(ys), smooth)

            if "val" in k:
                _type = "--"
            elif "test" in k:
                _type = ":"
            else:
                _type = "-"

            ax.plot(xs, ys, _type, label=k)
        ax.grid(True)
        ax.set_xlabel("Epochs")
        ax.set_ylabel("Metrics")
        ax.set_title(
            kwargs.get("title", None), fontdict=kwargs.get("fontdict", {"size": 20})
        )
        if kwargs.get("log", False):
            ax.semilogy()
        ax.legend()
        if _show:
            plt.show()

    def history(self, k):
        return [v for _, v in getattr(self, k)]

    def plot_epochs(self, keys: list = None, ax=None, **kwargs):
        _show = True if ax is None else False
        if ax is None:
            sz = 8, 6
            fig, ax = plt.subplots(figsize=kwargs.get("figsize", sz))
        avgs = defaultdict(list)
        keys = self.logged if keys is None else keys

        from tqdm import trange

        if isinstance(keys, str):
            key_pattern = keys
            keys = [key for key in self.logged if re.search(key_pattern, key)]
        xs = []
        for epoch in trange(-100, self.n_epochs + 1):
            for k in keys:
                items = takewhile(
                    lambda x: epoch - 1 <= x.pos < epoch,
                    dropwhile(
                        lambda x: (epoch - 1 > x.pos or x.pos > epoch), getattr(self, k)
                    ),
                )
                items = list(items)
                if items == []:
                    continue
                xs.append(epoch)
                avgs[k].append(np.mean([v for pos, v in items]))
            xs = sorted(set(xs))

        for k in avgs:
            if "val" in k:
                _type = "--"
            elif "test" in k:
                _type = ":"
            else:
                _type = "-"
            if len(avgs[k]) != len(xs):
                logger.info(
                    f"metric {k} was not fully recorded. Plotting final epochs using last recorded value"
                )
                avgs[k].extend([avgs[k][-1]] * (len(xs) - len(avgs[k])))
            ax.plot(
                xs,
                avgs[k],
                _type,
                label=k,
            )
        ax.grid(True)
        ax.set_xlabel("Epochs")
        ax.set_ylabel("Metrics")
        ax.set_title(
            kwargs.get("title", None), fontdict=kwargs.get("fontdict", {"size": 20})
        )
        if kwargs.get("log", False):
            ax.semilogy()
        plt.legend()
        if _show:
            plt.show()

    def report_avgs(self, epoch, return_avgs=True, end="\n"):
        avgs = {}
        for k in self.logged:
            avgs[k] = np.mean(
                [v for pos, v in getattr(self, k) if epoch - 1 <= pos < epoch]
            )
        self.report_metrics(epoch, end=end, **avgs)
        avgs = {f"epoch_{k}": v for k, v in avgs.items()}
        if self.wandb_logging:
            wandb.log({**avgs, "epoch": epoch})
        if self.mlflow_logging:
            mlflow.log_metrics(avgs, step=epoch)
        if return_avgs:
            return avgs

    def report_metrics(self, pos, **report):
        """Report training and validation metrics
        Required variables to be initialized before calling this function:
        1. start (time.time())
        2. n_epochs (int)

        Special kwargs:
        1. end - line ending after print (default newline)
        2. log - prefix info before printing summary

        Special argument:
        1. pos - position in training/testing process - float between 0 - n_epochs

        Usage:
        report_metrics(pos=1.42, train_loss=train_loss, validation_loss=validation_loss, ... )
        where each kwarg is a float
        """
        elapsed = time.time() - self.start
        end = report.pop("end", "\n")
        log = report.pop("log", "")
        log = log + ": " if log != "" else log
        elapsed = "  ({:.2f}s - {:.2f}s remaining)".format(
            time.time() - self.start, ((self.n_epochs - pos) / pos) * elapsed
        )
        current_iteration = f"EPOCH: {pos:.3f}  "
        if end == "\r":
            print(
                f"\r{log}{current_iteration}{info(report, self.precision)}{elapsed}",
                end="",
            )
        else:
            print(
                f"\r{log}{current_iteration}{info(report, self.precision)}{elapsed}",
                end=end,
            )

    def set_external_logging(self, project=None, hyper_parameters=None):
        if project is not None:
            if "/" in project:
                project, name = project.split("/")
            self.wandb_logging = True
            wandb.init(project=project, name=name, config=hyper_parameters)
        else:
            self.wandb_logging = False

        if mlflow and mlflow.active_run():
            self.mlflow_logging = True
        else:
            self.mlflow_logging = False

        self.train_step = 0
        self.validation_step = 0

    def finish_run(self, **kwargs):
        if not kwargs.get("do_not_finish_wandb", False):
            if self.wandb_logging:
                wandb.finish()


try:
    import pytorch_lightning as pl
    from pytorch_lightning.callbacks.progress import ProgressBarBase

    class LightningReport(ProgressBarBase):
        def __init__(
            self,
            epochs,
            print_every=None,
            print_total=None,
            precision=4,
            old_report=None,
        ):
            super().__init__()
            self.enable = True
            self.epoch_ix = 0
            _report = old_report.report if old_report is not None else None
            self.report = Report(epochs, precision, _report)
            if print_every is not None:
                self.print_every = print_every
            else:
                if print_total is None:
                    if epochs < 11:
                        self.print_every = 1
                else:
                    self.print_every = epochs // print_total

        def disable(self):
            self.enable = False

        def on_epoch_end(self, trainer, pl_module):
            self.epoch_ix += 1
            if self.epoch_ix % self.print_every == 0:
                self.report.report_avgs(self.epoch_ix)

        def prefix_dict_keys(self, prefix, _dict):
            _d = {}
            for k in _dict:
                _d[f"{prefix}_{k}"] = _dict[k]
            return _d

        def on_train_batch_end(
            self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
        ):
            super().on_train_batch_end(
                trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
            )
            if isinstance(outputs, list):
                loss = float(trainer.progress_bar_dict["loss"])
                outputs = outputs[0][0]["extra"]
                outputs["loss"] = loss
            outputs = self.prefix_dict_keys("trn", outputs)
            self.px = self.epoch_ix + ((1 + batch_idx) / self.total_train_batches)
            self.report.record(self.px, end="\r", **outputs)

        def on_validation_batch_end(
            self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
        ):
            super().on_train_batch_end(
                trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
            )
            loss = float(trainer.progress_bar_dict["loss"])
            self.px = self.epoch_ix + ((1 + batch_idx) / self.total_val_batches)
            outputs = self.prefix_dict_keys("val", outputs)
            self.report.record(self.px, end="\r", **outputs)

        def __getattr__(self, attr, **kwargs):
            return getattr(self.report, attr, **kwargs)

    __all__ += ["LightningReport", "pl"]
except Exception as e:
    ...


def moving_average(a, n=3):
    b = np.zeros_like(a)
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    _n = len(b) - n
    b[-_n - 1 :] = ret[(n - 1) :] / n
    return b


def save_torch_model_weights_from(model, fpath, verbose=True):
    "from model to fpath"
    torch.save(model.state_dict(), fpath)
    if verbose:
        fsize = os.path.getsize(fpath) >> 20
        logger.opt(depth=1).log("INFO", f"Saved weights of size ~{fsize} MB to {fpath}")


def load_torch_model_weights_to(model, fpath, device=None, verbose=True):
    "to model from fpath"
    if not device:
        model.load_state_dict(torch.load(fpath))
    else:
        model.load_state_dict(torch.load(fpath, map_location=device))
    if verbose:
        logger.opt(depth=1).log("INFO", f"Loaded weights from {fpath} to given model")


def detach(i):
    if isinstance(i, dict):
        for k, v in i.items():
            i[k] = detach(v)
        return i
    elif isinstance(i, (list, tuple)):
        return [detach(j) for j in i]
    else:
        return i.cpu().detach()


def cat_with_padding(tensors, mode="constant", value=-100):
    """
    Concatenates a list of tensors with padding to make them compatible along 0th dimension.

    Args:
        tensors (list of torch.Tensor): List of tensors to be concatenated.
        mode (str, optional): The padding mode. Default is "constant".
        value (int, optional): The padding value for "constant" mode. Default is -100.

    Returns:
        torch.Tensor: Concatenated tensor with padding to match dimensions.

    Note:
        - All tensors should have the same dimension except for the stacking dimension.
        - Padding is added to match the dimensions of the largest tensor in the specified stacking dimension.
    """
    assert all(
        [tensors[0].ndim == t.ndim for t in tensors[1:]]
    ), "All tensors should have the same number of dimensions"
    sizes = [(t.size()[1:]) for t in tensors]
    full_size = [max(s) for s in list(zip(*sizes))]
    if full_size == []:
        # Given tensors are just scalars
        return torch.stack(tensors)

    def make_padding(current_size, target_size):
        o = ()
        for curr_d, targ_d in zip(current_size[::-1], target_size[::-1]):
            diff_d = targ_d - curr_d
            o = o + (0, diff_d)
        return o

    out = torch.cat(
        [
            F.pad(x, make_padding(x.size()[1:], full_size), mode=mode, value=value)
            for x in tensors
        ],
        dim=0,
    )
    return out


def clean_ipython_hist():
    # Code in this function mainly copied from IPython source
    if not "get_ipython" in globals():
        return
    ip = get_ipython()
    user_ns = ip.user_ns
    ip.displayhook.flush()
    pc = ip.displayhook.prompt_count + 1
    for n in range(1, pc):
        user_ns.pop("_i" + repr(n), None)
    user_ns.update(dict(_i="", _ii="", _iii=""))
    hm = ip.history_manager
    hm.input_hist_parsed[:] = [""] * pc
    hm.input_hist_raw[:] = [""] * pc
    hm._i = hm._ii = hm._iii = hm._i00 = ""


def clean_tb():
    # h/t Piotr Czapla
    if hasattr(sys, "last_traceback"):
        traceback.clear_frames(sys.last_traceback)
        delattr(sys, "last_traceback")
    if hasattr(sys, "last_type"):
        delattr(sys, "last_type")
    if hasattr(sys, "last_value"):
        delattr(sys, "last_value")


def clean_gpu_mem():
    clean_tb()
    clean_ipython_hist()
    gc.collect()
    torch.cuda.empty_cache()


def get_latest_checkpoint(directory, prefix="checkpoint-"):
    """Gets the latest checkpoint directory in the given directory.
    Args:
    directory: The directory to search for checkpoint directories.
    Returns:
    The path to the latest checkpoint directory, or None if no checkpoint directories
    were found.
    """
    import glob
    import os

    checkpoint_directories = sorted(
        glob.glob(os.path.join(directory, f"{prefix}*")),
        key=lambda x: int(x.split("/")[-1].replace(prefix, "")),
    )
    if not checkpoint_directories:
        return None
    return checkpoint_directories[-1]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/markups.ipynb.

# %% auto 0
__all__ = [
    "AD",
    "Config",
    "isnamedtupleinstance",
    "unpack",
    "AttrDict",
    "pretty_json",
    "read_json",
    "write_json",
    "write_jsonl",
    "read_jsonl",
    "read_yaml",
    "write_yaml",
    "read_xml",
    "write_xml",
]

import hashlib

# %% ../nbs/markups.ipynb 2
import json
from collections.abc import Mapping
from json import JSONEncoder
from typing import Union

import jsonlines
import xmltodict
import yaml

from .icecream import ic
from .loader import BB, L, np, pd
from .logger import *
from .paths import *
from .thinc_parser.parser import Config


# %% ../nbs/markups.ipynb 3
def _default(self, obj):
    return getattr(obj.__class__, "to_json", _default.default)(obj)


_default.default = JSONEncoder().default
JSONEncoder.default = _default


def isnamedtupleinstance(x):
    _type = type(x)
    bases = _type.__bases__
    if len(bases) != 1 or bases[0] != tuple:
        return False
    fields = getattr(_type, "_fields", None)
    if not isinstance(fields, tuple):
        return False
    return all(type(i) == str for i in fields)


def unpack(obj):
    if isinstance(obj, dict):
        return {key: unpack(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [unpack(value) for value in obj]
    elif isnamedtupleinstance(obj):
        return {key: unpack(value) for key, value in obj._asdict().items()}
    elif isinstance(obj, tuple):
        return tuple(unpack(value) for value in obj)
    else:
        return obj


def hash_tensor(tensor):
    try:
        tensor_str = tensor.cpu().detach().numpy().tobytes()
    except:
        ...
    hash_obj = hashlib.sha256(tensor_str)
    return "ID:#" + hash_obj.hexdigest()[:6]


def hash_pandas_dataframe(input):
    try:
        from pandas.util import hash_pandas_object

        h = hash_pandas_object(input, index=True).values
        return "ID:#" + hashlib.sha256(h).hexdigest()[:6]
    except:
        return "ID:#<uncomputable>"


class AttrDict(object):
    """
    Utility class to interact with a dictionary as if it were an object. `AD` is an alias to this class

    FEATURES:
    0. Access and modify keys (including nested keys) as if they were object attributes, supporting tab-completion.
       Example: `self.key1.key2[0].key3`
    1. Keys and values are recursively converted to AttrDict instances.
    2. Pretty-print the dictionary using `self.pretty()`.
    3. Convert the entire structure to a regular dictionary at any time using `self.to_dict()`.
    3. Recursively remove keys using `self.drop(key)` from a JSON object.
    4. Apply a function to all values at all levels using `map`.

    GOTCHAS:
    1. All integer keys are implicitly converted to strings due to the enforced `self.key` format.
    2. You can still use `self[int]`, but this internally converts the integer to a string.

    METHODS:
    - `items()`: Return the items of the AttrDict as key-value pairs.
    - `keys()`: Return the keys of the AttrDict.
    - `values()`: Return the values of the AttrDict.
    - `update(dict)`: Update the AttrDict with key-value pairs from another dictionary.
    - `get(key, default=None)`: Get the value associated with a key, with an optional default value.
    - `__iter__()`: Allow iteration over the keys of the AttrDict.
    - `__len__()`: Return the number of keys in the AttrDict.
    - `__repr__()`: Return a string representation of the AttrDict.
    - `__dir__()`: List the keys of the AttrDict as attributes.
    - `__contains__(key)`: Check if a key exists in the AttrDict.
    - `__delitem__(key)`: Delete a key from the AttrDict.
    - `map(func)`: Apply a function to all values in the AttrDict.
    - `drop(key)`: Recursively remove a key and its values from the AttrDict.
    - `to_dict()`: Convert the AttrDict and its nested structure to a regular dictionary.
    - `pretty(print_with_logger=False, *args, **kwargs)`: Pretty-print the AttrDict as JSON.
    - `__eq__(other)`: Compare the AttrDict with another dictionary for equality.
    - `find_address(key, current_path="")`: Find and return all addresses (paths) of a given key in the AttrDict.
    - `summary(current_path='', summary_str='', depth=0, sep='\t')`: Generate a summary of the structure and values in the AttrDict.
    - `write_summary(to, **kwargs)`: Write the summary to a file or stream.
    - `fetch(addr)`: Retrieve a value at a specified address (path).

    PARAMETERS:
    - `data` (dict, optional): Initial data to populate the AttrDict.

    USAGE:
    - Create an AttrDict instance by providing an optional initial dictionary, and then access and manipulate its contents as if they were object attributes.

    EXAMPLE:
    ```python
    my_dict = {'name': 'John', 'age': 30, 'address': {'city': 'New York', 'zip': '10001'}}
    attr_dict = AttrDict(my_dict)
    print(attr_dict.name)  # Access values like attributes
    attr_dict.address.city = 'Los Angeles'  # Modify nested values
    ```
    """

    forbidden = set(":,'\"}{")

    def __init__(self, *args, given_input_to_ad=None, **kwargs):
        given_input_to_ad = {} if given_input_to_ad is None else given_input_to_ad
        if len(args) == 1 and isinstance(args[0], (Mapping, AttrDict)):
            given_input_to_ad = args[0]
            args = {}
        else:
            _args = dict(ic.io(*args)) if len(args) > 0 else {}
            args = {}
            for k, v in _args.items():
                if any(c in self.forbidden for c in k):
                    assert isinstance(
                        v, (dict, AttrDict)
                    ), f"Input `{v}` can't be a list"
                    given_input_to_ad = {**v, **given_input_to_ad}
                else:
                    args = {**{k: v}, **args}

        given_input_to_ad = {**kwargs, **given_input_to_ad, **args}
        for name, value in given_input_to_ad.items():
            setattr(self, str(name), self._wrap(value))

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def _wrap(self, value):
        if isinstance(value, (L, tuple, list, set, frozenset)):
            value = type(value)([self._wrap(v) for v in value])
            if isinstance(value, (list, L)):
                value = L(value)
            return value
        else:
            return (
                AttrDict(given_input_to_ad=value) if isinstance(value, dict) else value
            )

    __getitem__ = (
        lambda self, x: AttrDict({_x: self[_x] for _x in x})
        if isinstance(x, (list, L))
        else getattr(self, str(x))
    )
    __setitem__ = lambda self, k, v: setattr(self, str(k), self._wrap(v))

    def update(self, dict):
        for k, v in dict.items():
            self[k] = v

    def get(self, key, default=None):
        key = str(key)
        return self[key] if key in self else default

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self.keys())

    def __repr__(self):
        return f"\n```↯ AttrDict ↯\n{self.summary()}\n```\n"

    def __dir__(self):
        return self.__dict__.keys()

    def __contains__(self, key):
        key = str(key)
        return key in self.__dict__.keys()

    def __delitem__(self, key):
        key = str(key)
        del self.__dict__[key]

    def map(self, func):
        for k in dir(self):
            v = self[k]
            if isinstance(v, AttrDict):
                v.map(func)
            elif isinstance(v, (L, tuple, list, set, frozenset)):
                v = [_v.map(func) if isinstance(_v, AttrDict) else func(_v) for _v in v]
            else:
                v = func(v)
            self[k] = v

    def drop(self, key):
        if key in self:
            del self[key]
        for k in dir(self):
            v = self[k]
            if isinstance(v, AttrDict):
                v.drop(key)
            if isinstance(v, (L, tuple, list, set, frozenset)):
                v = [_v.drop(key) for _v in v if isinstance(_v, AttrDict)]

    def to_dict(self):
        d = {}
        for k in dir(self):
            v = self[k]
            if isinstance(v, AttrDict):
                v = v.to_dict()
            if isinstance(v, (L, tuple, list, set, frozenset)):
                v = [_v.to_dict() if isinstance(_v, AttrDict) else _v for _v in v]
            d[k] = v
        return d

    def pretty(self, print_with_logger=False, *args, **kwargs):
        pretty_json(
            self.to_dict(), print_with_logger=print_with_logger, *args, **kwargs
        )

    def __eq__(self, other):
        return AttrDict(given_input_to_ad=other).to_dict() == self.to_dict()

    def find_address(self, key, current_path=""):
        addresses = []
        for k in self.keys():
            if current_path:
                new_path = f"{current_path}.{k}"
            else:
                new_path = k

            if k == key:
                addresses.append(new_path)

            if isinstance(self[k], AttrDict):
                addresses.extend(self[k].find_address(key, new_path))

            elif isinstance(self[k], (L, tuple, list, set, frozenset)):
                for i, item in enumerate(self[k]):
                    if isinstance(item, AttrDict):
                        addresses.extend(item.find_address(key, f"{new_path}.{i}"))
        return addresses

    def summary(self, current_path="", depth=0, sep="  ", max_items=10):
        def format_path(path, key):
            return f"{path}.{key}" if path else key

        def format_item(key, item, path, depth, sep):
            import numpy as np
            import pandas as pd
            import torch

            if isinstance(item, (pd.DataFrame,)):
                return f"{sep * depth}{key} - {type(item).__name__} - shape {item.shape} - columns {item.columns} - {hash_pandas_dataframe(item)}\n"
            if isinstance(item, AttrDict) or hasattr(item, "keys"):
                item = AttrDict(**item)
                return f"{sep*depth}{key}\n" + item.summary(path, depth + 1, sep)
            elif isinstance(item, (list, tuple, set, frozenset, L)):
                return summarize_collection(key, item, path, depth + 1, sep)
            elif isinstance(item, (torch.Tensor, np.ndarray)):
                if isinstance(item, np.ndarray):
                    item = torch.tensor(item)
                return f"{sep * depth}{key} - {item} - {hash_tensor(item)}\n"

            else:
                if isinstance(item, (int, float, complex, str)):
                    is_multiline = False
                    if isinstance(item, str):
                        is_multiline = "\n" in item
                        if len(item) > 200:
                            item = item[:100] + "..." + item[-100:]
                        if is_multiline:
                            _item = item.split("\n")
                            _item = "\n".join([f"{sep*(depth+1)}{l}" for l in _item])
                            item = f"↓\n{sep*(depth+1)}```\n{_item}\n{sep*(depth+1)}```"
                    multiline = "" if not is_multiline else "Multiline "
                    return f"{sep * depth}{key} - {item} ({multiline}{type(item).__name__})\n"
                else:
                    return f"{sep * depth}{key} - {type(item).__name__}\n"

        def summarize_collection(key, collection, path, d, s):
            summary_str = f"{s * (d - 1)}{key}\n"
            for i, item in enumerate(collection):
                item_path = format_path(path, i)
                if i < max_items:
                    summary_str += format_item(i, item, item_path, d, s)
                else:
                    summary_str += (
                        f"{s*d}... {len(collection) - max_items} more items ...\n"
                    )
                    break
            return summary_str

        summary_str = ""
        for ix, key in enumerate(self.keys()):
            if ix >= max_items:
                summary_str += (
                    f"{sep*depth} ... {len(self.keys()) - max_items} more keys ...\n"
                )
                break
            new_path = format_path(current_path, key)
            summary_str += format_item(key, self[key], new_path, depth, sep)
        return summary_str

    def print_summary(self, **kwargs):
        from builtins import print

        print(self.summary(**kwargs))

    def write_summary(self, to, **kwargs):
        writelines(self.summary(**kwargs).split("\n"), to)

    def fetch(self, addr):
        if isinstance(addr, (list, L)):
            return L([self.fetch(_addr) for _addr in addr])

        o = self
        for p in addr.split("."):
            try:
                o = o[int(p)]
            except:
                o = o[p]
        return o


AD = AttrDict
AD.dict = AD.to_dict


def pretty_json(
    i, fpath=None, indent=4, print_with_logger=True, return_as_string=False
):
    def set_default(obj):
        if isinstance(obj, (set, BB, L)):
            return list(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, P):
            return str(obj)

    assert isinstance(i, (dict, list))
    i = unpack(i)
    dump = json.dumps(i, indent=indent, default=set_default)
    if fpath:
        makedir(parent(fpath))
        logger.info(f"Dumped a pretty file to {fpath}")
        with open(fpath, "w") as f:
            json.dump(i, f, indent=indent, default=set_default)
            return
    if print_with_logger:
        return logger.opt(depth=1).log("DEBUG", f"\n{dump}")
    else:
        if return_as_string:
            return dump
        print(dump)


# %% ../nbs/markups.ipynb 8
def read_json(fpath):
    import json

    with open(fpath, "r") as f:
        return json.load(f)


def write_json(obj, fpath, silent=False):
    from datetime import date, datetime

    def set_default(obj):
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        raise TypeError

    if not silent:
        logger.opt(depth=1).log("DEBUG", f"Dumping json to {fpath}")
    with open(fpath, "w") as f:
        json.dump(obj, f, indent=4, default=set_default)
    return P(fpath)


# %% ../nbs/markups.ipynb 10
def write_jsonl(items, dest, mode="a"):
    makedir(parent(dest))
    with jsonlines.open(dest, mode) as writer:
        writer.write_all(items)
        if mode == "a":
            Info(f"Appended {len(items)} items to {dest}")
        elif mode == "w":
            Info(f"Wrote {len(items)} jsons to {dest}")


def read_jsonl(file):
    return [json.loads(line) for line in readlines(file, silent=True)]


# %% ../nbs/markups.ipynb 11
def read_yaml(file):
    with open(file, "r") as stream:
        try:
            return yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)


def write_yaml(content, fpath):
    with open(fpath, "w") as outfile:
        yaml.dump(content, outfile, default_flow_style=False)


# %% ../nbs/markups.ipynb 12
def read_xml(file_path: Union[str, P]) -> AttrDict:
    "Read xml data as a dictionary"
    with open(str(file_path)) as xml_file:
        data = AttrDict(given_input_to_ad=xmltodict.parse(xml_file.read()))
    return data


def write_xml(data: Union[AttrDict, dict], file_path: Union[str, P]):
    makedir(parent(file_path))
    "convert a dictionary to xml"
    with open(file_path, "w") as xml_file:
        data = data.to_dict() if isinstance(data, AttrDict) else data
        assert isinstance(data, dict), "Function only supports dicts for now"
        data = xmltodict.unparse(data, pretty=True)
        xml_file.write(data)


# %% ../nbs/markups.ipynb 13
Config = Config

if __name__ == "__main__":
    assert AD({}) == {}

    p = 1
    q = {"a": 10}
    s = AD(p, q, a=20, b=30)
    assert s == {"a": 20, "b": 30, "p": 1, "q": {"a": 10}}

    s = AD(p, **q, b=30)
    assert s == {"a": 10, "b": 30, "p": 1}

    p = {"b": 222}
    assert AD(p, {"a": 2, "b": 2}, {"bb": 2, "a": 3, "b": 3}, a=20, b=30) == {
        "a": 2,
        "b": 2,
        "bb": 2,
        "p": {"b": 222},
    }

    d1 = {"c": 1}
    d2 = {"c": 2}
    d3 = {"c": 3}
    l = [1, 2, 3]
    assert AD(l, {"c": 100}, d1, d2, d3, d1={"c": 10}, d2={"c": 20}).to_dict() == {
        "l": [1, 2, 3],
        "d1": {"c": 1},
        "d2": {"c": 2},
        "c": 100,
        "d3": {"c": 3},
    }

    a = 20
    b = 30
    assert AD(a, b).to_dict() == {"a": 20, "b": 30}

    p = 10
    q = {"a": 1}
    assert AD(p, q) == {"p": 10, "q": {"a": 1}}
    assert AD(p, {"a": 1}) == {"p": 10, "a": 1}

    x = {"a": 1, "b": 2, "c": 3}
    assert AD(x).to_dict() == {"x": {"a": 1, "b": 2, "c": 3}}
    assert AD(**x).to_dict() == {"a": 1, "b": 2, "c": 3}
